#
#  @file   B2buaCall.py
#  @author Francis Moreau
#  date    Mon May 22 2007
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

from SipCallEngineExtending import PySipMessage
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import *

from paraxip.sipua.internal.SipCallEngineSMCall import *
from paraxip.sipua.internal.ConnectionStatus import *
#import paraxip.sipua.internal.SipMessageHelperFunctions
from paraxip.sipua.internal.ParameterInfo import ParameterInfoDB

from B2buaStates import *

import re

#
# b2bua state machine
#

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

class B2buaCall(SipCallEngineSMCall):
  
  # Valid parameter names for this Call
  PARAM_B2BUA_OUTBOUND_PROXY_HOST = 'b2bua.outbound.proxy.host'
  PARAM_B2BUA_OUTBOUND_PROXY_PORT = 'b2bua.outbound.proxy.port'
  PARAM_B2BUA_OUTBOUND_REQUESTURI_HOST = 'b2bua.outbound.requestURI.host'
  PARAM_B2BUA_OUTBOUND_REQUESTURI_PORT = 'b2bua.outbound.requestURI.port'
  PARAM_B2BUA_OUTBOUND_REQUESTURI_USERPART = 'b2bua.outbound.requestURI.userpart'

  # Provide parameter information for this Call 
  @staticmethod
  def getParameterInfoDB():

    parameterInfoDB = ParameterInfoDB()

    # Optional outbound proxy address:port
    parameterInfoDB.addParameterInfoForName(
        B2buaCall.PARAM_B2BUA_OUTBOUND_PROXY_HOST,
        type=str,
        mandatory=False)
    parameterInfoDB.addParameterInfoForName(
        B2buaCall.PARAM_B2BUA_OUTBOUND_PROXY_PORT,
        type=int,
        mandatory=False)

    # Optional outbound requestURI address:port
    #  To use instead of the caller-provided request URI address:port
    #  (ie when we're not set as the outbound proxy by the caller)
    parameterInfoDB.addParameterInfoForName(
        B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_HOST,
        type=str,
        mandatory=False)
    parameterInfoDB.addParameterInfoForName(
        B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_PORT,
        type=int,
        mandatory=False)

    # Internal - Optional outbound requestURI user part
    #  To use instead of the caller-provided request URI user part
    parameterInfoDB.addParameterInfoForName(
        B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_USERPART,
        type=str,
        mandatory=False)

    return parameterInfoDB

  #@override
  def __init__(self):
    SipCallEngineSMCall.__init__(self)
    
    self.setSMCallLogger()
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.__init__")

    # Generic ReINVITED state
    self._reinvitedState = None

    self.addStates()

    ###
    ### PER CALL MEMBERS
    ###
    
    self.callerDialogId = None
    self.calleeDialogId = None
    self.declarePerCallMembers('callerDialogId', 'calleeDialogId')
   
    self.callerSdp = None
    self.calleeSdp = None
    self.declarePerCallMembers('callerSdp', 'calleeSdp')
    
    ###
    ### Call State Variables
    ###
    
    self.callerLegStatus = ConnectionStatus.Unknown
    self.declarePerCallMembers('callerLegStatus') 
    
    self.calleeLegStatus = ConnectionStatus.Unknown
    self.declarePerCallMembers('calleeLegStatus')    
    
    self.bInvitedWithSdp = False
    self.declarePerCallMembers('bInvitedWithSdp')
    
    self.bCalleeAnswered = False
    self.declarePerCallMembers('bCalleeAnswered')
    
    #@tag Marshall and remove if possible
    self.uiCalleeSipErrorCode = None
    self.declarePerCallMembers('uiCalleeSipErrorCode')
    
    #@tag Marshall and remove if possible
    self.uiRejectCallerWithInternalSipErrorCode = None    
    self.declarePerCallMembers('uiRejectCallerWithInternalSipErrorCode')
    
    #@tag Marshall and remove if possible
    self.strHangupWithInternalReasonHeader = None
    self.declarePerCallMembers('strHangupWithInternalReasonHeader')
    
    self.callerInitialInvite = None # Initial Caller invite msg
    self.declarePerCallMembers('callerInitialInvite')
    
    self.dialogIdInitiatingDisconnect = None
    self.declarePerCallMembers('dialogIdInitiatingDisconnect')
    
    ### PER CALL MEMBERS (Call State Variables) END
    
    ### PER CALL MEMBERS END

  #@override
  def getName(self):
    return 'b2bua'

  def getReInvitedState(self):
    return self._reinvitedState

  def setSMCallLogger(self):
    self.setLogger(CallLogger("netborder.b2bua.call.py"))
  
  def addStates(self):
    """
    @note addState(name, state) is defined in PySipCallEngineSM.hpp
    """
    dialingCalleeState = DialingCalleeState(self)

    # IDLE state
    idleState = IdleState(self, dialingCalleeState.getName())
    self.setInitialStateName(idleState.getName())
    self.setInitialState(idleState)
    
    # DIALING_CALLEE state
    self.addState(dialingCalleeState.getName(), dialingCalleeState)
    
    # EARLY_MEDIA state
    earlyMediaState = EarlyMediaState(self)
    self.addState(earlyMediaState.getName(), earlyMediaState)
    
    # ANSWERING_CALLER state
    answeringCallerState = AnsweringCallerState(self)
    self.addState(answeringCallerState.getName(), answeringCallerState)
    
    # CALLER_CALLEE_CONNECTED state
    callerCalleeConnectedState = CallerCalleeConnectedState(self)
    self.addState(callerCalleeConnectedState.getName(),
                  callerCalleeConnectedState)
    
    # DISCONNECTING state
    disconnectingState = DisconnectingState(self)
    self.addState(disconnectingState.getName(), disconnectingState)
    
    # REINVITED state
    self._reinvitedState = ReinvitedState(self)
    self._reinvitedState.setSuccessStateName(callerCalleeConnectedState.getName())
    self._reinvitedState.setFailureStateName(callerCalleeConnectedState.getName())
    self._reinvitedState.setDisconnectStateName(disconnectingState.getName())
    self.addState(self._reinvitedState.getName(), self._reinvitedState)

    # DISCONNECTED state
    disconnectedState = DisconnectedState(self)
    self.addFinalState(disconnectedState.getName(), disconnectedState)
  
  #@override
  def initialize(self, in_refApplication, in_refGlobalContext, in_szCallId):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.initialize")
    
    SipCallEngineSMCall.initialize(self,
                                   in_refApplication,
                                   in_refGlobalContext,
                                   in_szCallId)

    return True
  
  #@override
  def reset(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.reset")

    SipCallEngineSMCall.reset(self)
  
  #@override
  def addInDialog(self, in_szDialogId, in_sipCallEngineSMIn):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.addInDialog")

    SipCallEngineSMCall.addInDialog(self, in_szDialogId, in_sipCallEngineSMIn)
    
    # Save the caller dialog id
    self.callerDialogId = in_szDialogId
  
  #@override
  def addOutDialog(self, in_szDialogId, in_sipCallEngineSMOut):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.addOutDialog")

    SipCallEngineSMCall.addOutDialog(self, in_szDialogId, in_sipCallEngineSMOut)
    
    # Save callee dialog id
    self.calleeDialogId = in_szDialogId
    
  def scenarioName(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.scenarioName")

    return "Netborder B2BUA"
  
  def notifyLoopDetected(self):
    """
    Virtual function overwritten in children than want to avoid using the
    uiRejectCallerWithInternalSipErrorCode and strHangupWithInternalReasonHeader
    variables.
    """
    
    self.uiRejectCallerWithInternalSipErrorCode = 482 # Loop Detected
    self.strHangupWithInternalReasonHeader = \
      "Loop Detected (A Request-Uri to a B2BUA must be different than " \
      "its address/port)"
  
  def callCallee(self, in_sdp):

    # By default, the request URI to use to reach callee is the same than the 
    #   one receive from the caller
    strCalleeRequestUriScheme = self.callerInitialInvite.getRequestUriScheme();
    strCalleeRequestUriUserPart = \
        self.callerInitialInvite.getUserPart("INVITE");
    strCalleeRequestUriHost = self.callerInitialInvite.getRequestUriHost();
    iCalleeRequestUriPort = self.callerInitialInvite.getRequestUriPort();
    strCalleeRequestUriParams = self.callerInitialInvite.getRequestUriParams()

    #Bug 4932 - Override the user part if instructed to
    if B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_USERPART in self.config:
      strCalleeRequestUriUserPart = \
          self.config[B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_USERPART]

    # Override host and port if instructed to
    if B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_HOST in self.config:
      strCalleeRequestUriHost = \
          self.config[B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_HOST]

    if B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_PORT in self.config:
      iCalleeRequestUriPort = \
          self.config[B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_PORT]

    # Construct Request URI to send to callee
    strCalleeRequestURI = "%s:%s@%s:%d%s" % (
      strCalleeRequestUriScheme, \
      strCalleeRequestUriUserPart, \
      strCalleeRequestUriHost, \
      iCalleeRequestUriPort, \
      strCalleeRequestUriParams)

    # Reject the incoming call if the resulting request URI is pointing 
    #   to ourself
    if (self.isCallerRequestUriIsLooping(strCalleeRequestURI)):
      self.notifyLoopDetected()
      return False

    strOutboundProxyURI = None
    if (B2buaCall.PARAM_B2BUA_OUTBOUND_PROXY_HOST in self.config):
      strOutboundProxyURI = "sip:%s:%d" % \
          (self.config[B2buaCall.PARAM_B2BUA_OUTBOUND_PROXY_HOST],
           self.config[B2buaCall.PARAM_B2BUA_OUTBOUND_PROXY_PORT])
      
    # Get an initial SIP INVITE message
    inviteMsg = PySipMessage()
    if (not self.globalContext.constructInviteMessage(inviteMsg,
                                                      self.getCallId(),
                                                      strCalleeRequestURI,
                                                      in_sdp,
                                                      strOutboundProxyURI)):
      # Error logged by constructInviteMessage
      return False

    self.mergeCommonInviteHeaders(self.callerInitialInvite, inviteMsg)

    # Send the SIP INVITE message
    self.globalContext.makeOutCall(inviteMsg)

    self.calleeLegStatus = ConnectionStatus.Connecting
    return True

  def mergeCommonInviteHeaders(self, in_sourceInvite, in_destInvite):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.mergeCommonInviteHeaders")
    
    # Copy the request-uri params from the initial SIP INVITE message
    in_destInvite.mergeRequestUriParams(in_sourceInvite)

    # Copy the From header from the initial SIP INVITE message
    in_destInvite.mergeNameAddr("From", in_sourceInvite)
    
    # Copy the To header from the initial SIP INVITE message
    # (bugfix for bugzilla::3894)
    in_destInvite.mergeNameAddr("To", in_sourceInvite)

    # copy the custom headers from the initial SIP INVITE message
    in_destInvite.mergeCustomHeaders(in_sourceInvite)

  def getAnswerMessage(self, in_sourceSipMessage=None):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.getAnswerMessage")
    
    sipMessage = PySipMessage()
    
    # TODO: Check everywhere and find a final solution for this => getInDialog
    #       and getOutDialog can fail in a race condition where the Dialog has
    #       been released by Resiprocate but the Call did not received the
    #       DialogToCallTerminatedEvent from the Dialog yet
    # It's possible that this Dialog Id has already been released but the Call
    # did not received the DialogToCallTerminatedEvent from this Dialog yet
    inDialogSM = self.getInDialog(self.callerDialogId)
    if (inDialogSM is None):
      PyLoggingFunctions.log_info(self.getLogger(),
               "getInDialog failed because dialog id %s "
               "does not exist or has been released", self.callerDialogId)
      return None

    inDialogSM.constructAnswerMessage(sipMessage)
    
    if (in_sourceSipMessage is not None):
      # copy the custom headers from the initial SIP INVITE message
      sipMessage.mergeCustomHeaders(in_sourceSipMessage)

    return sipMessage
  
  def getRejectMessage(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.getRejectMessage")
    
    sipMessage = PySipMessage()
    
    # It's possible that this Dialog Id has already been released but the Call
    # did not received the DialogToCallTerminatedEvent from this Dialog yet
    inDialogSM = self.getInDialog(self.callerDialogId)
    if (inDialogSM is None):
      PyLoggingFunctions.log_info(self.getLogger(),
               "getInDialog failed because dialog id %s "
               "does not exist or has been released", self.callerDialogId)
      return 603, None

    inDialogSM.constructRejectMessage(sipMessage)
    
    # Get the sip reject response code
    uiRejectResponseCode = 603 # Decline
    if (self.uiRejectCallerWithInternalSipErrorCode is not None):
      uiRejectResponseCode = self.uiRejectCallerWithInternalSipErrorCode
    elif (self.uiCalleeSipErrorCode is not None):
      uiRejectResponseCode = self.uiCalleeSipErrorCode
    
    # Add custom header here
    # assert sipMessage.addCustomHeader("header name", "header value")
    
    return uiRejectResponseCode, sipMessage

  def isCallerRequestUriIsLooping(self, in_szRequestUri):
    """
    Check if the caller RequestUri is pointing on ourself (which could create
    a infinite loop when configured as a Local Outbound Proxy)
    """
    stackHostNameAndPort = self.globalContext.getHostName() + \
                           ":" + \
                           str(self.globalContext.getSIPStackListeningPort())
    stackIPAddressAndPort = self.globalContext.getSIPStackIPAddress() + \
                            ":" + \
                            str(self.globalContext.getSIPStackListeningPort())

    PyLoggingFunctions.log_debug(self.getLogger(),
        "isCallerRequestUriIsLooping: Looking for '%s' or '%s' in '%s'" % 
        (stackHostNameAndPort, stackIPAddressAndPort, in_szRequestUri))

    compiledOnHostName = re.compile(stackHostNameAndPort, re.IGNORECASE)
    compiledOnIPAddress = re.compile(stackIPAddressAndPort)
    searchedOnHostName = \
      compiledOnHostName.search(in_szRequestUri)
    searchedOnIPAddress = \
      compiledOnIPAddress.search(in_szRequestUri)
    return (searchedOnHostName is not None) or (searchedOnIPAddress is not None)

  def relayDialogToCallInfoEvent(self, in_event):
    """
    Relays received DTMF SIP info messages between connected caller and callee.
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaCall.relayDialogToCallInfoEvent")

    if (self.callerLegStatus != ConnectionStatus.Connected or
       self.calleeLegStatus != ConnectionStatus.Connected):
        PyLoggingFunctions.log_info(self.getLogger(),
                                    "Not relaying received SIP INFO message"
                                    " since call is not connected yet.")
        return False

    sipMsg = in_event.getSipMessage()
    if sipMsg is None:
        PyLoggingFunctions.log_warn(self.getLogger(),
                                    "Failed to retrieved INFO sip message")
        return False
    if sipMsg.hasContents() is not True:
        PyLoggingFunctions.log_info(self.getLogger(),
                                    "No Content in receved SIP INFO message."
                                    " Discarding it.")
        return False

    if (sipMsg.getContentType() != "application" or
        (sipMsg.getContentSubType() != "dtmf-relay" and
         sipMsg.getContentSubType() != "dtmf")):
        PyLoggingFunctions.log_info(self.getLogger(),
                                    "Discarding unsupported content %s:%s" %
                                    (sipMsg.getContentType(),
                                     sipMsg.getContentSubType()))
        return False
    # At this point we're sure that we're in connected state and that the
    # received SIP INFO message is a DTMF relay message, so forward it...
    forward_dialog_id = None
    if in_event.getDialogId() == self.callerDialogId:
        forward_dialog_id = self.calleeDialogId
    elif in_event.getDialogId() == self.calleeDialogId:
        forward_dialog_id = self.callerDialogId
    else:
        PyLoggingFunctions.log_warn(self.getLogger(),
                                    "Unknown event dialog id %s" %
                                    (in_event.getDialogId()))
        return False

    if forward_dialog_id is None:
        PyLoggingFunctions.log_warn(self.getLogger(),
                                    "Invalid forward dialog id.")
        return False

    PyLoggingFunctions.log_info(self.getLogger(),
                                "Forwarding received sip INFO with content %s"
                                " to dialog with id %s" %
                                (sipMsg.getBodyText(),
                                 forward_dialog_id))

    return SipCallEngineSMCall.sendInfoMsgToDialog(self, forward_dialog_id,
                                                   sipMsg.getContentType(),
                                                   sipMsg.getContentSubType(),
                                                   sipMsg.getBodyText())
## !!MIRROR FILE, DO NOT MODIFY!!
