

from netborderPyInfra import Logger 
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions

from paraxip.sipua.ui.model.cpaOM import CpaOM 

import re
import string

cpaStatsLogger = Logger('netborder.cpa.stats')

def _url_decode(in_str):
  ##  URL encoding steps are as follows:
  ##          - The ASCII characters 'a' through 'z', 'A' thru 'Z', '0' through '9', 
  ##           and ".", "-", "*", "_" remain the same 
  ##          - The space character (ASCII 20) is converted into a + sign
  ##          - The remaining characters are represented by 3-character strings 
  ##           which begin with the percent sign, "%xy", where xy is the two-digit 
  ##           hexadecimal representation of the UTF-8 encoded character
  ##          Decoding involves theses steps in reverse
  ##           XXX Encoding steps above are not exactly the same as for SIP URLs 
  ##           (see Section 25.1 of RFC 3261), but it doesn't matter for decoding

  pos = 0
  size = len(in_str)
  out_strDecoded = ""

  # Encoded URLa uses only base ASCII characters
  while pos < size:
    c = in_str[pos]
    if c == '%':
      if((pos + 2 < size)  and (in_str[pos+1] in string.hexdigits) and
         (in_str[pos+2] in string.hexdigits)):
        # The next two chars are hex digits: convert them to ASCII
        try:
          n = str(unichr(int(in_str[(pos+1):(pos+3)],16)))
          out_strDecoded += n
          pos += 2
        except:
          # exception occured, just skip the current hexadecimal
          # character and continue
          pass
          
      else:
        out_strDecoded += '%'
    else:
      out_strDecoded += c
    
    pos += 1
      
  return out_strDecoded


def _appendStatsToList(in_cpaOM, in_list, in_key):
  
  if in_cpaOM.isOM(in_key):
    item = in_cpaOM.getOM(in_key)
    if not isinstance(item, basestring):
      item = 'PyNotStr_' + str(item)
    in_list.append(item)
  else:
    in_list.append("")

def WriteToCsv(in_cpaOM, in_strCallId, in_strCallerRequestURL):
  """
   Write CPA stats using the following CSV format:
     - Paraxip Call-id  
     - Date of the call 
     - Campaign DBID (empty if it does not exist) 
     - Campaign Name (empty if it does not exist)
     - Phone number  (exclude public access code)
     - CPA Result 
     - Time Dialing 
     - Time Connected 
     - Time CPA Completed 
     - Time Queued (if relevant, empty if it doesn't exist) 
     - Time Connected to agent
  """
  
  cpaStatsList = []
  
  # Call-id
  in_cpaOM.addOM(CpaOM.OM_CALL_ID, in_strCallId)
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CALL_ID)
  
  # Date
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_DATE)
  
  # Campaign ID
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_CAMPAIGN_ID)
  
  # Campaign Name
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_CAMPAIGN_NAME)
  
  # Phone number
  strCallerRequestURL = _url_decode(in_strCallerRequestURL)
  phoneNumberRegEx = re.compile('sip:(\S+)@')
  if strCallerRequestURL:
    m = phoneNumberRegEx.match(strCallerRequestURL)
    if m:
      cpaStatsList.append(m.group(1))
    else:
      cpaStatsList.append("")
  else:
    cpaStatsList.append("")
  
  # CPA result
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_RESULT)
  
  # Time Dialing
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_TIME_DIALING)

  # Time Connected
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_TIME_CONNECTED)

  # Time CPA completed
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_TIME_CPA_COMPLETED)

  # Time callee was queued before talking to an agent
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_TIME_QUEUED)

  # Time agent was connected
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_CPA_TIME_AGENT_CONNECTED)
  
  # New special call information field
  _appendStatsToList(in_cpaOM, cpaStatsList, CpaOM.OM_SPECIAL_CALL_INFORMATION)

  # Fast way to concatenate string list using "," as separator
  strCpaStats = ','.join(cpaStatsList)
  
  # Finally logging the CSV line
  PyLoggingFunctions.log_info(cpaStatsLogger, strCpaStats)
  
  
def test_module():
  from paraxip.sipua.ui.model.cpaOM import CpaOM
  cpaOM = CpaOM()
  WriteToCsv(cpaOM, 'test_call_ID', 'test_caller_url')

if __name__ == '__main__':
  test_module()
## !!MIRROR FILE, DO NOT MODIFY!!
