import paraxip.sipua.CheckImport
paraxip.sipua.CheckImport.CheckImport(__file__)

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import Logger,TraceScope

from SipCallEngineExtending import PySipOptionsRequestResult

fileScopeLogger = Logger('netborder.cpa.__init__')

# NOTE do not import anything locally without aliases, as the imports will be visible in .sipua.user.cpa for subsequent calls
from paraxip.sipua.user.NetBorderApp import NetBorderApp as _NetBorderApp

###############################################################################
# Call Analyzer Engine (ie CPA UAS process)

class Engine(_NetBorderApp):

  def onInviteGetCall(self, in_sipMessage):
    __pychecker__ = 'unusednames=in_sipMessage'
    """
    USER CONFIGURATION 
      Returns a tuple ([SM class:Class], [config:dict]) to tell which scenario \
      to run, and with what configuration.
      Must make sure that the named class is known (ie was imported)

    """
    from paraxip.sipua.user.cpa.CpaUASCall import CpaUASCall
    return (CpaUASCall,
            {
              #Nothing here; part of the configuration is received from
              #the B2BUA as a SIP nessage (see InDialogToCallCalledEvent)
              #in CpaUASStates.py.
            })

###############################################################################
# Call Analyzer Service base class

class CallAnalyzerApp(_NetBorderApp):
  
  # Valid parameter names for this App
  PARAM_NCA_ENDOFGREETING_ENABLE = 'nca.EndOfGreeting.enable'
  PARAM_NCA_REQUIRE_CPD_ON_TO_PERFORM_CPA = \
      'nca.RequireCpdOnToPerformCpa'
  PARAM_NCA_ACCEPT_CALLS_WHEN_CPA_NOT_AVAIL = \
      'nca.AcceptCallsWhenCPANotAvailable'
  # Bug 4932: Call subclass selection via prefix
  PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX = 'nca.ExpectPhoneNumberPrefix'

  # Aliases for Call properties - call properties have precedence over these
  PARAM_NCA_PRECONNECT_TIMEOUT_MS = \
      'nca.PreConnectTimeout.ms'
  PARAM_NCA_POSTCONNECT_TIMEOUT_MS = \
      'nca.PostConnectTimeout.ms'
  PARAM_NCA_ALLOW_POST_CONNECT_TIMEOUT_DELAYING = 'nca.PostConnectTimeout.AllowDelaying'
  PARAM_NCA_END_OF_GREETING_TIMEOUT_MS = \
      'nca.EndOfGreetingTimeout.ms'
  PARAM_NCA_HUMAN_THRESHOLD = \
      'nca.HumanThreshold'
  PARAM_NCA_MACHINE_THRESHOLD = \
      'nca.MachineThreshold'
  PARAM_NCA_FAX_THRESHOLD = \
      'nca.FaxThreshold'
  PARAM_NCA_ENGINE_HOST = \
      'nca.EngineHost'
  PARAM_NCA_ENGINE_PORT = \
      'nca.EnginePort'
  PARAM_NCA_PROVISIONAL_RELAYED_UPON_183 = \
      'nca.provisionalRelayedUpon183'
  PARAM_NCA_RELAY_PROVISIONAL_WHILE_WAITING_FOR_RESULT = 'nca.relayProvisionalWhileWaitingForResult'
  # Added Parameters to control preconnect thresholds
  PARAM_NCA_SIT_PERMANENT_THRESHOLD = \
      'nca.SitPermanentThreshold'
  PARAM_NCA_SIT_TEMPORARY_THRESHOLD = \
      'nca.SitTemporaryThreshold'
  PARAM_NCA_BUSY_THRESHOLD = \
      'nca.BusyThreshold'
  PARAM_NCA_REORDER_THRESHOLD = \
      'nca.ReorderThreshold'
  

  # Provide parameter information for this App
  @staticmethod
  def getParameterInfoDB():
    from paraxip.sipua.internal.ParameterInfo import ParameterInfoDB
    parameterInfoDB = ParameterInfoDB()
    
    # Parameters that must be set explicitly (mandatory, no defaults)
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_ENDOFGREETING_ENABLE,
        type=bool)
    
    # Parameters with default values
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_REQUIRE_CPD_ON_TO_PERFORM_CPA,
        type=bool,
        defaultValue=False)
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_ACCEPT_CALLS_WHEN_CPA_NOT_AVAIL,
        type=bool,
        defaultValue=False)
    # Bug 4932: Call subclass selection via prefix
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX,
        type=bool,
        defaultValue=False)

    # Parameters that are aliases to Call parameters
    from paraxip.sipua.user.cpa.CpaCall import CpaCall 
    # Ending comma for aliases is to create a tuple
    # http://docs.python.org/reference/expressions.html#grammar-token-expression_list
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_PRECONNECT_TIMEOUT_MS,
        type=int,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_PRECONNECT_TIMEOUT_MS,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_POSTCONNECT_TIMEOUT_MS,
        type=int,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_POSTCONNECT_TIMEOUT_MS,))
    
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_ALLOW_POST_CONNECT_TIMEOUT_DELAYING,
        type=bool,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_ALLOW_POST_CONNECT_TIMEOUT_DELAYING,))
    
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_HUMAN_THRESHOLD,
        type=float,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_HUMAN,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_MACHINE_THRESHOLD,
        type=float,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_ANSWERING_MACHINE,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_FAX_THRESHOLD,
        type=float,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_FAX,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_ENGINE_HOST,
        type=str,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_ENGINE_HOST,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_ENGINE_PORT,
        type=int,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_ENGINE_PORT,))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_PROVISIONAL_RELAYED_UPON_183,
        type=int,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_PROVISIONAL_RELAYED_UPON_183,))

    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_RELAY_PROVISIONAL_WHILE_WAITING_FOR_RESULT,
        type=bool,
        mandatory=False,
        aliases=(CpaCall.PARAM_CPA_RELAY_PROVISIONAL_WHILE_WAITING_FOR_RESULT,))
     
    parameterInfoDB.addParameterInfoForName(
      CallAnalyzerApp.PARAM_NCA_SIT_PERMANENT_THRESHOLD,
      type=float,
      mandatory=False,
      aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_SIT_PERMATENT,))
    parameterInfoDB.addParameterInfoForName(
      CallAnalyzerApp.PARAM_NCA_SIT_TEMPORARY_THRESHOLD,
      type=float,
      mandatory=False,
      aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_SIT_TEMPORARY,))
    parameterInfoDB.addParameterInfoForName(
      CallAnalyzerApp.PARAM_NCA_BUSY_THRESHOLD,
      type=float,
      mandatory=False,
      aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_BUSY,))
    parameterInfoDB.addParameterInfoForName(
      CallAnalyzerApp.PARAM_NCA_REORDER_THRESHOLD,
      type=float,
      mandatory=False,
      aliases=(CpaCall.PARAM_CPA_RESULT_THRESHOLD_REORDER,))

    from paraxip.sipua.user.cpa.CpaEamdCall import CpaEamdCall 
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerApp.PARAM_NCA_END_OF_GREETING_TIMEOUT_MS,
        type=int,
        mandatory=False,
        aliases=(CpaEamdCall.PARAM_CPA_EAMD_TIMEOUT_MS,))

    return parameterInfoDB

  ##########################################################################

  def __init__(self):
    super(CallAnalyzerApp, self).__init__()
    # Application level configuration
    self.__config = None
    # Effective user part (phone number) to reach callee
    self._calleeUserPart = None
    return

  #@abstract
  def getNonCpaCallCls(self):
    assert False
    
  #@override
  # Application level configuration
  def getConfig(self):
    return self.__config

  #@override
  def onStartup(self, in_refGlobalContext):
    # Call base class method
    assert super(CallAnalyzerApp, self).onStartup(in_refGlobalContext)

    # Get app config from defaults + config URL
    import paraxip.sipua.internal.call_config_loader as call_config_loader
    self.__config = call_config_loader.getAppConfig(self.__class__)
    if self.__config is None:
      # Error already logged by function
      return False

    # Verify that config is valid for app (errors logged by function if any)
    fileName = call_config_loader.getConfigURLForApp(self.__class__)
    return self.__class__.getParameterInfoDB().validateConfig(\
        self.__config, fileName)

  # helper to extract custom headers from the initial INVITE message from callee
  def _addAllCustomHeadersToCPAStats(self, config, in_sipINVITEMessage):
    #Extract custom headers from SIP INVITE's header.
    dict_inviteSIPMsgCustomHeaders = {}
    in_sipINVITEMessage.getCustomHeaders(dict_inviteSIPMsgCustomHeaders)
    from paraxip.sipua.user.cpa.CpaCall import CpaCall
    config[CpaCall.PARAM_CPA_STATS_CUSTOM_HEADERS] = \
      dict_inviteSIPMsgCustomHeaders

  #@override
  def onInviteGetCall(self, in_sipMessage):
    """
    USER CONFIGURATION 
      Returns a tuple ([SM class:Class], [config:dict]) to tell which scenario \
      to run, and with what configuration.
    """
    
    # Check for unsupported property values
    if self.getConfig()[\
        CallAnalyzerApp.PARAM_NCA_ACCEPT_CALLS_WHEN_CPA_NOT_AVAIL]:
      errorMsg = 'Setting parameter %s to true is not yet supported' % \
        CallAnalyzerApp.PARAM_NCA_ACCEPT_CALLS_WHEN_CPA_NOT_AVAIL
      PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
      return self.rejectCall(500, errorMsg)

    # The goal of the onInviteGetCall method is to select a Call subclass 
    #   in fonction of the incoming INVITE
    callCls = None
    config = {}
    
    doCPA = True
    if self.getConfig()[\
        CallAnalyzerApp.PARAM_NCA_REQUIRE_CPD_ON_TO_PERFORM_CPA]:
      doCPA = (in_sipMessage.getRequestLine().find("cpd=on") != -1)

    doEndOfGreeting = self.getConfig()[\
        CallAnalyzerApp.PARAM_NCA_ENDOFGREETING_ENABLE]

    # Extract user part (phone number) of Request URI
    strUserPart = in_sipMessage.getUserPart("INVITE");

    ###########################################################################
    # Bug 4932: Override Call subclass selection based on phone number 
    # (User Part of Request URI) prefix
    if self.getConfig()[\
        CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX]:

      # Need at least 4 digits for the prefix + 1 digit to dial
      if len(strUserPart) < 5:
        errorMsg = '4 digit prefix + phone number must be provided when %s ' \
          'parameter is set to true' % \
          (CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX,)
        PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
        return self.rejectCall(500, errorMsg)

      # Prefix must start with '99'
      if strUserPart[:2] != '99':
        errorMsg = '4 digit prefix must start with "99" when %s parameter' \
          ' is set to true' % \
          (CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX,)
        PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
        return self.rejectCall(500, errorMsg)

      # Interpret 3rd (feature code) and 4th (feature value) digit 
      #   (index 2 and 3)
      # Only 0 and 1 supported as feature code
      if strUserPart[2] not in ('0','1'):
        errorMsg = 'Only "0" and "1" supported after prefix of 99 when %s ' \
            'parameter is set to true' % \
          (CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX,)
        PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
        return self.rejectCall(500, errorMsg)

      if strUserPart[2] == '0':
        # no-op
        pass

      if strUserPart[2] == '1':
        # Answering Machine detection mode
        if strUserPart[3] not in ('0','1', '2'):
          errorMsg = 'Only "0", "1" and "2" supported after prefix of 991 ' \
              'when %s parameter is set to true' % \
            (CallAnalyzerApp.PARAM_NCA_EXPECT_PHONE_NUMBER_PREFIX,)
          PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
          return self.rejectCall(500, errorMsg)

        if strUserPart[3] == '0':
          # Answering machine detection disabled
          doCPA = False
          doEndOfGreeting = False
        if strUserPart[3] == '1':
          # Answering machine detection enabled, End Of Greeting disabled
          doCPA = True
          doEndOfGreeting = False
        if strUserPart[3] == '2':
          # Answering machine detection enabled, End Of Greeting enabled
          doCPA = True
          doEndOfGreeting = True

      # Strip prefix by telling to use as user part the original user part minus 
      #   the first 4 digits
      self._calleeUserPart = strUserPart[4:]
    else:
      # Not using prefix: Callee User Part is the same as Caller
      self._calleeUserPart = strUserPart
      if len(strUserPart) == 0:
          errorMsg = 'Must provide a user part (phone number) in requests to NCA'
          PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
          return self.rejectCall(500, errorMsg)

    ###########################################################################
    # Select the Call class to use

    if doCPA:
      from paraxip.sipua.user.cpa.CpaCall import CpaCall
      from paraxip.sipua.user.cpa.CpaEamdCall import CpaEamdCall 

      self._addAllCustomHeadersToCPAStats(config, in_sipMessage)

      # Bug 2359: Dynamically change pre-connect timeout on per call basis
      dictCustomHeaders = {}
      in_sipMessage.getCustomHeaders(dictCustomHeaders)
      if CpaCall.CPA_PRECONNECT_TIMEOUT_CUSTOM_HEADER_NAME in dictCustomHeaders:
        strHeaderValue = dictCustomHeaders[\
            CpaCall.CPA_PRECONNECT_TIMEOUT_CUSTOM_HEADER_NAME]
        try:
          # Header values are strings - must be convertible to an int
          # Ring Timeout is in seconds - convert to ms
          config[CpaCall.PARAM_CPA_PRECONNECT_TIMEOUT_MS] = 1000 * \
              int(strHeaderValue)
        except ValueError:
          errorMsg = 'Header %s must have an integer value, "%s" provided' % \
              (CpaCall.CPA_PRECONNECT_TIMEOUT_CUSTOM_HEADER_NAME, strHeaderValue)
          PyLoggingFunctions.log_error(fileScopeLogger, errorMsg)
          return self.rejectCall(500, errorMsg)
      
      if doEndOfGreeting:
        callCls = CpaEamdCall 
      else:
        callCls = CpaCall

    else:      
      callCls = self.getNonCpaCallCls()

    # Push the user part (number) to use to reach callee into the config
    from paraxip.sipua.user.redirect.RedirectCall import RedirectCall
    if callCls == RedirectCall:
      config[RedirectCall.PARAM_REDIRECT_CONTACT_USERPART] = \
          self._calleeUserPart
    else:
      from paraxip.sipua.user.b2bua.B2buaCall import B2buaCall
      config[B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_USERPART] = \
          self._calleeUserPart

    return self.updateAndValidateConfigForAppAndCall(config, self, callCls)

###############################################################################
# Call Analyzer Service, targeted as being an outbound proxy, using b2bua Call
#   for non-cpa calls

class CallAnalyzerAsOutboundProxy(CallAnalyzerApp):
  
  #@override
  def getNonCpaCallCls(self):
    from paraxip.sipua.user.b2bua.B2buaCall import B2buaCall
    return B2buaCall

###############################################################################
# Call Analyzer Service, NOT targeted as being an outbound proxy, using redirect for non-cpa calls

class CallAnalyzerGenesysOCS(CallAnalyzerApp):
  
  PARAM_NCA_RELAY_SERVER_HOST = 'nca.RelayServerHost'
  PARAM_NCA_RELAY_SERVER_PORT = 'nca.RelayServerPort'
  PARAM_NCA_FORWARD_SIP_OPTIONS_TO_RELAY_SERVER = \
      'nca.ForwardSipOptionsToRelayServer'

  #@override
  # Provide parameter information for this App
  @staticmethod
  def getParameterInfoDB():
    from paraxip.sipua.user.b2bua.B2buaCall import B2buaCall
    from paraxip.sipua.user.redirect.RedirectCall import RedirectCall

    parameterInfoDB = CallAnalyzerApp.getParameterInfoDB()
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerGenesysOCS.PARAM_NCA_RELAY_SERVER_HOST,
        type=str,
        mandatory=False,
        aliases=(\
          B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_HOST,
          RedirectCall.PARAM_REDIRECT_CONTACT_HOST
          ))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerGenesysOCS.PARAM_NCA_RELAY_SERVER_PORT,
        type=int,
        mandatory=False,
        aliases=(\
          B2buaCall.PARAM_B2BUA_OUTBOUND_REQUESTURI_PORT,
          RedirectCall.PARAM_REDIRECT_CONTACT_PORT
          ))
    parameterInfoDB.addParameterInfoForName(
        CallAnalyzerGenesysOCS.PARAM_NCA_FORWARD_SIP_OPTIONS_TO_RELAY_SERVER,
        type=bool,
        defaultValue=False
        )

    return parameterInfoDB

  #@override
  def getNonCpaCallCls(self):
    from paraxip.sipua.user.redirect.RedirectCall import RedirectCall
    return RedirectCall

  #@override
  def onOptionsRequest(self, in_sipMessage):
    """
      Called on reception of OPTIONS request outside a dialog
      @returns SipOptionsRequestResult telling whether OPTIONS forwarding must
               be performed, and to which address and port if it's the case
    """
    __pychecker__ = 'unusednames=t,in_sipMessage'
    t = TraceScope(fileScopeLogger,
                                    "NetBorderApp.onOptionsRequest")

    doForward = self.getConfig()[\
        CallAnalyzerGenesysOCS.PARAM_NCA_FORWARD_SIP_OPTIONS_TO_RELAY_SERVER]

    forwardHost = self.getConfig()[\
        CallAnalyzerGenesysOCS.PARAM_NCA_RELAY_SERVER_HOST]

    forwardPort = self.getConfig()[\
        CallAnalyzerGenesysOCS.PARAM_NCA_RELAY_SERVER_PORT]


    return PySipOptionsRequestResult(
        doForward,
        forwardHost,
        forwardPort)

###############################################################################

# remove aliases no longer needed
del _NetBorderApp
## !!MIRROR FILE, DO NOT MODIFY!!
