#
#  @file   RedirectStates.py
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2008 Paraxip Technologies.
#  All Rights Reserved.
#
import re

from netborderPyInfra import TraceScope
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions

from SipCallEngineExtending import InitialEvent
from paraxip.sipua.internal.SipCallEngineState import SipCallEngineState
from paraxip.sipua.internal.SipCallEngineEvent import \
    InDialogToCallCalledEvent, DialogToCallSdpOfferedEvent, \
    DialogToCallTerminatedEvent


# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

#
# Call states
#

#
# Base class
#

class RedirectStates(SipCallEngineState):
  def __init__(self, in_stateMachine):
    SipCallEngineState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "RedirectStates.__init__")

#
# Idle state
#

class IdleState(RedirectStates):
  def __init__(self, in_stateMachine):
    RedirectStates.__init__(self, in_stateMachine)
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "IdleState._init__")
  
  def getName(self):
    return "IDLE"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)

    out_pystrNewStateName.set(self.getName())

    if (isinstance(in_event, InitialEvent)):
      return True

    elif (isinstance(in_event, InDialogToCallCalledEvent)):

      # Default some contact URI fields from the request URI received from the 
      #   caller
      strContactUriScheme = in_event.getSipMessage().getRequestUriScheme();
      strContactUriUserPart = \
          in_event.getSipMessage().getUserPart("INVITE");
      strContactUriParams = in_event.getSipMessage().getRequestUriParams()

      # Bug 3382: put explicit transport URI parameter in Contact for redirects
      #  if not UDP (default)
      if strContactUriParams.lower().find('transport') == -1:
        # No indication of transport - specify it following transport on 
        #   which INVITE was received (if not the default UDP)
        strTransport = in_event.getSipMessage().getReceivedTransportName()
        if strTransport != 'UDP':
          strContactUriParams += ';transport=%s' % (strTransport,) 

      from paraxip.sipua.user.redirect.RedirectCall import RedirectCall

      #Bug 4932 - Override the user part if instructed to
      if RedirectCall.PARAM_REDIRECT_CONTACT_USERPART in self._stateMachine.config:
        strContactUriUserPart = \
            self._stateMachine.config[RedirectCall.PARAM_REDIRECT_CONTACT_USERPART]

      # Redirect using
      # sip:[phoneNumber]@[contact.host]:[contact.port][;request_params] as 
      #   the contact
      contact = 'sip:%s@%s:%d%s' % (strContactUriUserPart, 
        self._stateMachine.config[RedirectCall.PARAM_REDIRECT_CONTACT_HOST], 
        self._stateMachine.config[RedirectCall.PARAM_REDIRECT_CONTACT_PORT],
        strContactUriParams)
      if (not self._stateMachine.redirectInDialog(
            in_event.getDialogId(), [contact])):
        return False

      return True # Do not change state, DialogToCallTerminatedEvent coming up

    elif (isinstance(in_event, DialogToCallSdpOfferedEvent)):
      return True # Stay in same state

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      out_pystrNewStateName.set("DISCONNECTED")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Disconnected state
#
# TODO this state is the same for all scenarios

class DisconnectedState(RedirectStates):
  def __init__(self, in_stateMachine):
    RedirectStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DisconnectedState.__init__")
  
  def getName(self):
    return "DISCONNECTED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    # No event should be received in this state
    PyLoggingFunctions.log_warn(self.getLogger(),
             "%s::processEvent received an invalid event %s",
             self.__class__.__name__, in_event.__class__.__name__)
    
    return True
## !!MIRROR FILE, DO NOT MODIFY!!
